/* 
 * @(#) $Header: /usr/src/mash/repository/mash/mash-1/codec/ldct/ldct-huff.cc,v 1.2 1997/08/09 03:39:38 tecklee Exp $ (LBL) 
 */

#include "config.h"
#include <stdio.h>
#include <string.h>

#define HUFFSTRINGS
#include "ldct-ht.h"
/*
 * Convert a binary string to an integer.
 */
int
btoi(const char* s)
{
	int v = 0;
	while (*s) {
		v <<= 1;
		v |= *s++ - '0';
	}
	return (v);
}

#define EOB LDCT_EOB

struct huffcode ldct_dchuff[] = {
#include "ldct-dchuff.h"
{  0, 0, }, 
};
struct huffcode ldct_layer0[] = {
#include "ldct-layer0.h"
{  0, 0, },
};
struct huffcode ldct_layer1[] = {
#include "ldct-layer1.h"
{  0, 0, },
};
struct huffcode ldct_layer2[] = {
#include "ldct-layer2.h"
{  0, 0, },
};


struct huff {
	char* name;
	huffcode* codes;
	int width;
};

struct huff hc[] = {
	{ "ldct_dchuff", ldct_dchuff, 10 },
	{ "ldct_layer0", ldct_layer0, 10 },
	{ "ldct_layer1", ldct_layer1, 10 },
	{ "ldct_layer2", ldct_layer2, 10 },
	{ 0, 0 }
};

struct hufftab {
	int maxlen;
	int *prefix;
};

/*
 * Build a direct map huffman table from the array of codes given.
 * We build a prefix table such that we can directly index
 * table with the k-bit number at the head of the bit stream, where
 * k is the max-length code in the table.  The table tells us
 * the value and the actual length of the code we have.
 * We need the length so we can tear that many bits off
 * the input stream.  The length and value are packed as
 * two 16-bit quantities in a 32-bit word.  The value is stored
 * in the upper 16-bits, so when we right-shift it over,
 * it is automatically sign-extended.
 */
void
huffbuild(hufftab& ht, const huffcode* codes)
{
	int i = 0;
	int maxlen = 0;
	for (; codes[i].str != 0; ++i) {
		int v = strlen(codes[i].str);
		if (v > maxlen)
			maxlen = v;
	}
	int size = 1 << maxlen;
	if (size > 1 << 20) {
		fprintf(stderr,
			"ldct-huff: max length codeword too long (%d)\n",
			maxlen);
		exit(1);
	}
		
	/*
	 * Build the direct-map lookup table.
	 */
	ht.prefix = new int[size];
	ht.maxlen = maxlen;

	/*
	 * Initialize states to illegal, and arrange
	 * for max bits to be stripped off the input.
	 */
	for (i = 0; i < size; ++i)
		ht.prefix[i] = (SYM_ILLEGAL << 16) | maxlen;

	for (i = 0; codes[i].str != 0; ++i) {
		int codelen = strlen(codes[i].str);
		int nbit = maxlen - codelen;
		int code = btoi(codes[i].str) << nbit;
		int map = (codes[i].val << 16) | codelen;
		/*
		 * The low nbit bits are don't cares.
		 * Spin through all possible combos.
		 */
		for (int n = 1 << nbit; --n >= 0; ) {
			if ((code | n) >= size)
				abort();
			ht.prefix[code | n] = map;
		}
	}
}

void 
gen_hte(huff* h)
{
	int esc = LDCT_ESC;
	int esclen = 0;
	/*
	 * find escape code: if we don't find it, fill unused
	 * slots with an illegal entry (-1,0)
	 */
	for (huffcode* p0 = h->codes; p0->str != 0; ++p0) {
		if (p0->val == LDCT_ESC) {
			esc = btoi(p0->str);
			esclen = strlen(p0->str);
			break;
		}
	}


	int n = 1 << h->width;
	int* isesc = new int[n];
	printf("const struct huffent hte_%s[%d] = {\n", h->name, n);
	int i;
	for (i = 0; i < n; ++i) {
		huffcode* p;
		isesc[i] = 0;
		for (p = h->codes; p->str != 0; ++p) {
			if (i == p->val) {
				printf("\t{ %d, %d },\n",
				       btoi(p->str), strlen(p->str));
				break;
			}
		}
		if (p->str == 0) {
			/* escape */
			printf("\t{ %d, %d },\n", esc, esclen);
			isesc[i] = 1;
		}
	}
	printf("};\n");

	printf("const int hte_%s_esc[%d] = {\n", h->name, n);
	for (i = 0; i < n; ++i) 
		printf("%d, ", isesc[i]);
	
	printf("\n};\n");
	delete isesc;
}

void
gen_htd(huff* h)
{
	hufftab ht;
	huffcode* p = h->codes;
	huffbuild(ht, p);
	printf("const int htd_%s_width = %d;\n", h->name, ht.maxlen);
	printf("const int htd_%s[] = {", h->name);
	int n = 1 << ht.maxlen;
	for (int i = 0; i < n; ++i)
		printf("%s%d,", (i & 7) ? " " : "\n\t",
		       ht.prefix[i]);
	printf("\n};\n");
}

int main()
{
	printf("#include \"p64-huff.h\"\n");
	for (huff* h = hc; h->name != 0; ++h) {
		gen_hte(h);
		gen_htd(h);
	}
        return 0;
}

