#
# Copyright (c) 1998 Regents of the University of California.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#       This product includes software developed by the Computer Systems
#       Engineering Group at Lawrence Berkeley Laboratory.
# 4. Neither the name of the University nor of the Laboratory may be used
#    to endorse or promote products derived from this software without
#    specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# ------------
#
# Filename: new-tcl-expand.tcl
#   -- Created on Fri Jul 24 1998
#   -- Author: Yatin Chawathe <yatin@cs.berkeley.edu>
#
# Description:
#
#
#  @(#) $Header: /usr/src/mash/repository/mash/mash-1/tcl-expand.tcl,v 1.15 1998/07/27 22:16:55 yatin Exp $
#




# expand the file "name" on the command line by subtituting "source"
# and "import" commands by the corresponding files

proc expand_file name {
	exit_if_url $name

	# check if the file has already been loaded
	global loaded
	if [info exists loaded($name)] {
		return
	}

	puts stderr "expanding {$name}"
	set loaded($name) 1
	if { $name != "stdin" } {
		if [catch {set f [open $name r]} msg] {
			puts stderr "*** error: can't open file $name"
			exit 1
		}
	} else {
		set f stdin
	}

	while 1 {
		if { [gets $f line] < 0 } {
			if { $f != "stdin" } {
				close $f
			}
			return
		}

		# check if the first non-space chars on the line are
		# "source" or "import"
		set keyword [string range [string trimleft $line] 0 5]

		# if you do not want this proc to expand a line containing
		# "source" or "import", add a backslash before the keyword
		if { $keyword == "source" || $keyword == "import" } {
			# Collapse whitespace before splitting line, so that
			# each extra space does not result in a {} list
			# element.
			
			regsub -all " +" $line " " newline
			set L [split [string trim $newline]]

			if { [llength $L] == 2 && [lindex $L 0] == "source" } {
				expand_file [lindex $L 1]
			} elseif { [lindex $L 0] == "import" } {
				set items [read_import_line $f $L]
				expand_imports $items
			} else {
				# this was a false alarm
				# it wasn't actually "source" or "import"
				puts $line
			}
		} else {
			puts $line
		}
		
	}
}




#
# Return 1 if the <i>name</i> is an http URL and 0 otherwise.
# An http URL is of the form http://<server and possibly port num>/
# Note that the trailing slash is important.
#
proc is_http_url { name } {
	if [regexp {([^:]+)://([^:/]+)(:([0-9]+))?(/.*)} $name url protocol \
			server x port trailingpath] {
		if { ![info exists protocol] } {
			return 0
		} else {
			return [string match http $protocol]
		}
	} else {
		return 0
	}
}


proc exit_if_url name {
	if [is_http_url $name] {
		puts stderr "exiting: retrieval of remote files ($name)\
				not yet supported"
		exit 1
	}
}


proc read_import_line { f L } {
	#
	# Check for backslash continuation
	#

	set items [lrange $L 1 end]

	while 1 {
		if { [lindex $items end] != "\\" } {
			return $items
		}
		set items [lrange $items 0 [expr [llength $items] - 2]]
		if { [gets $f line] < 0 } {
			return $items
		}
		set items "$items $line"
	}
}


proc expand_imports { items } {
	global importTable block init
	
	if [info exists block] {
		# an error had occurred previously; ignore all import commands
		return
	}
	
	if ![info exists init] {
		# we haven't initialized the import table yet! do it.
		set init 1
		if ![read_table] {
			puts stderr "...ignoring import commands"
			set block 1
		}
	}

	foreach item $items {
		puts stderr "importing $item"
		if ![info exists importTable($item)] {
			puts stderr "*** warning: no info for $item in\
					importTable"
			continue
		}
		foreach file $importTable($item) {
			expand_file $file
		}
	}
}


proc find_import_dirs {} {
	global env
	if [file exists importLocation] {
		set fd [open importLocation "r"]
		while 1 {
			if { [gets $fd line] < 0 } {
				close $fd 
				break
			}
			lappend list $line
		}
	} elseif [info exists env(TCLCL_IMPORT_DIRS)] {
		set list [split $env(TCLCL_IMPORT_DIRS) :]
	} else {
		set list .
	}
	set dirs ""
	foreach d $list {
		if [file readable $d/importTable] {
			lappend dirs $d
		}
	}
	return $dirs
}


proc read_dir dir {
	global importTable
	set f [open $dir/importTable "r"]
	while 1 {
		set l [gets $f]
		if [eof $f] {
			close $f
			break
		}
		set index [lindex $l 0]
		set fname [lindex $l 1]
		
		set fname [file join $dir $fname]
		if {[info exists importTable($index)]} {
			if {-1!=[lsearch -exact $importTable($index) $fname]} {
				continue
			}
		}
		lappend importTable($index) $fname
	}
	return 1
}


proc read_table { } {
	set dirs [find_import_dirs]
	if { $dirs == "" } {
		puts stderr "can't find an importTable in current directory"
		puts stderr "or in path from TCLCL_IMPORT_DIRS environment"
		return 0
	}
	foreach d $dirs {
		read_dir $d
	}
	return 1
}



# expand each filename on the command line by subtituting "source"
# and "import" commands by the corresponding files
foreach name $argv {
	expand_file $name
}

exit 0



