
#provide UDPServer
import UDPChannel

# UDPChannel that maintains list of unicast
# address/port pairs and optionally also a single mcast address/port; 
# all are sent data when 'send' is called.
# It automatically times out addrs that haven't been refreshed in
# 'timeout_' seconds (if timeout_ == 0 they persist forever); 
# the list can also be manually manipulated via 
# 'addAddr' and 'rmAddr' and 'timeoutAddrs'
#
Class UDPServer -superclass UDPChannel

# spec indicates the port number to listen on; additionally,
# if spec is a multicast addr, it is used in addition to unicast addrs
# as a location to send 'announce' messages
#
UDPServer public init { {spec ""} {mtu 1024} } {
    $self instvar replyToList_ mcastAddr_ port_ child_ timeout_

    set timeout_ 600 ; # default timeout is 10mins
    set replyToList_ ""
    set replyTimeList_ ""

    set port_ [lindex [split $spec "/"] 1]
	# addressblocks only allow odd ports -- emulate this
	if {$port_ % 2 == 1} {
		incr port_ -1
		set spec "[lindex [split $spec "/"] 0]/$port_"
	}

    # if spec includes a name, turn it into IP addr
    set firstchar [string index $spec 0]
    if [string match \[a-zA-Z\] $firstchar] {
	set n [lindex [split $spec "/"] 0]
	set p [lindex [split $spec "/"] 1]
	set s [gethostbyname $n]
	if { $s == "" } {
	    puts "cannot find address for '$n'"
	    exit
	}
	set spec $s/$p
    }

    if [in_multicast $spec] {
	$self next $spec $mtu
	set mcastAddr_ $spec
	set child_ [new UDPServerChild $self 255.255.255.255/$port_]
    } else {
	set mcastAddr_ -1
	set child_ [new UDPServerChild $self $spec]
    }

}

# add addr and/or refresh last update time
# for addrspec in replyToList_
UDPServer public addAddr {addrspec} {
    $self instvar replyToList_ replyTimeList_

    set l [lsearch $replyToList_ $addrspec]
    if {$l == -1} {
	lappend replyToList_ $addrspec
	lappend replyTimeList_ [clock seconds]
    } else {
	set replyTimeList_ [lreplace $replyTimeList_ $l $l [clock seconds]]
    }
}

# remove addr from replyTo list
UDPServer public rmAddr {addrspec} {
    $self instvar replyToList_ replyTimeList_

    set loc [lsearch $replyToList_ $addrspec]
    if {$loc != -1} {
	set replyToList_ [lreplace $replyToList_ $loc $loc]
	set replyTimeList_ [lreplace $replyTimeList_ $loc $loc]
    } else {
	puts "UDPServer: Tried to remove addr not in reply list."
    }
}

# remove any addr not updated in last <em>timeout<em> seconds
#
UDPServer public timeoutAddrs {timeout} {
    $self instvar replyToList_ replyTimeList_

    set t [clock seconds]
    set cnt 0
    foreach i $replyTimeList_ {
	if {[expr $t - $i] > $timeout} {
	    $self rmAddr [lindex $replyToList_ $cnt]
	}
	incr cnt
    }
}

UDPServer public announce {data} {
    $self instvar replyToList_ mcastAddr_ port_ timeout_

    if {$timeout_ > 0} {
	$self timeoutAddrs $timeout_
    }

    foreach i $replyToList_ {
	puts "$i: $data"
	set sender [new UDPChannel $i]
	$sender send "$data"
	delete $sender
    }
    if {$mcastAddr_ != -1} {
	$self send "$data"
    }

    # XXX backward compatability: for clients that listen on the port
    # number they send to (i.e., well-known rather than ephemeral)
    regsub -all "/" "$replyToList_" " " tmp
    array set addrsOnly $tmp
    foreach i [array names addrsOnly] {
	set sender [new UDPChannel $i/$port_]
	$sender send "$data"
	delete $sender
    }
}

# set new timeout value
UDPServer public timeout {t} {
    $self instvar timeout_
    set timeout_ $t
}

# Receive Stub: should be extended by subclasses
UDPServer private recv {addr port data len} {
    puts "Msg from $addr/$port \[$len\]: $data"
}


# -----------

# UDPServerChild is a UDPServer helper class: it simply monitors a
# unicast address and passes along received packets to the parent
# after adding the addresses to the replyTo list
Class UDPServerChild -superclass UDPChannel

#
UDPServerChild instproc init {parent spec {mtu 1024}} {
    $self next $spec $mtu
    $self instvar parent_
    set parent_ $parent
}

#
UDPServerChild instproc recv {addr port data size} {
    $self instvar parent_
    $parent_ addAddr $addr/$port
    $parent_ recv $addr $port "$data" $size
}

